-- File output functions for XPMC

include assoc.e
include globals.e
--include mdx.e
include vgm.e
include wav.e
--include ym.e


-- Output table data in WLA-DX format 
global function output_wla_table(sequence name, sequence tbl, integer canLoop, integer scaling, integer loopDelim)
	integer bytesWritten
	
	bytesWritten = 0
	
	if length(tbl[1]) then
		for i = 1 to length(tbl[1]) do
			printf(outFile, name & "_%d:", tbl[1][i])
			for j = 1 to length(tbl[2][i][2]) do
				if canLoop and j = length(tbl[2][i][2]) and length(tbl[2][i][3]) = 0 then
					if j > 1 then
						printf(outFile, ", $%02x", loopDelim)
					end if
					printf(outFile, {13, 10} & name & "_%d_loop:" & {13, 10}, tbl[1][i])
					printf(outFile, ".db $%02x, $%02x", {and_bits(tbl[2][i][2][j] * scaling, #FF), loopDelim})
					bytesWritten += 3
				elsif j = 1 then
					printf(outFile, {13, 10} & ".db $%02x", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 1
				else
					printf(outFile, ", $%02x", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 1
				end if
			end for
			if canLoop and length(tbl[2][i][3]) then
				if length(tbl[2][i][2]) then
					printf(outFile, ", $%02x", loopDelim)
					bytesWritten += 1
				end if
				printf(outFile, {13, 10} & name & "_%d_loop:" & {13, 10}, tbl[1][i])
				for j = 1 to length(tbl[2][i][3]) do
					if j = 1 then
						printf(outFile, ".db $%02x", and_bits(tbl[2][i][3][j] * scaling, #FF))
					else
						printf(outFile, ", $%02x", and_bits(tbl[2][i][3][j] * scaling, #FF))
					end if
					bytesWritten += 1
				end for
				printf(outFile, ", $%02x", loopDelim)
				bytesWritten += 1
			end if
			puts(outFile, {13, 10})
		end for
		puts(outFile, name & "_tbl:" & {13, 10})
		for i = 1 to length(tbl[1]) do
			printf(outFile, ".dw " & name & "_%d" & {13, 10}, tbl[1][i])
			bytesWritten += 2
		end for
		if canLoop then
			puts(outFile, name & "_loop_tbl:" & {13, 10})
			for i = 1 to length(tbl[1]) do
				printf(outFile, ".dw " & name & "_%d_loop" & {13, 10}, tbl[1][i])
				bytesWritten += 2
			end for
		end if
		puts(outFile, {13, 10})
	else
		puts(outFile, name & "_tbl:" & {13, 10})
		if canLoop then
			puts(outFile, name & "_loop_tbl:" & {13, 10})
		end if
		puts(outFile, {13, 10})
	end if
	return bytesWritten
end function


-- Output table data in m68k-as format 
global function output_m68kas_table(sequence name, sequence tbl, integer canLoop, integer scaling, integer longPointers)
	integer bytesWritten,ptrSize
	sequence ptrDecl
	
	bytesWritten = 0
	
	if longPointers then
		ptrDecl = "dc.l "
		ptrSize = 4
	else
		ptrDecl = "dc.w "
		ptrSize = 2
	end if
	
	if length(tbl[1]) then
		for i = 1 to length(tbl[1]) do
			--printf(outFile, ".globl " & name & "_%d", tbl[1][i])
			printf(outFile, name & "_%d:", tbl[1][i])
			for j = 1 to length(tbl[2][i][2]) do
				if canLoop and j = length(tbl[2][i][2]) and length(tbl[2][i][3]) = 0 then
					if j > 1 then
						puts(outFile, ", 0x80")
					end if
					--printf(outFile, CRLF & ".globl " & name & "_%d_loop" & CRLF, tbl[1][i])
					printf(outFile, CRLF & name & "_%d_loop:" & CRLF, tbl[1][i])
					printf(outFile, "dc.b 0x%02x, 0x80", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 3
				elsif j = 1 then
					printf(outFile, CRLF & "dc.b 0x%02x", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 1
				else
					printf(outFile, ", 0x%02x", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 1
				end if
			end for
			if canLoop and length(tbl[2][i][3]) then
				if length(tbl[2][i][2]) then
					puts(outFile, ", 0x80")
					bytesWritten += 1
				end if
				--printf(outFile, {13, 10} & ".globl " & name & "_%d_loop" & CRLF, tbl[1][i])
				printf(outFile, {13, 10} & name & "_%d_loop:" & CRLF, tbl[1][i])
				for j = 1 to length(tbl[2][i][3]) do
					if j = 1 then
						printf(outFile, "dc.b 0x%02x", and_bits(tbl[2][i][3][j] * scaling, #FF))
					else
						printf(outFile, ", 0x%02x", and_bits(tbl[2][i][3][j] * scaling, #FF))
					end if
					bytesWritten += 1
				end for
				puts(outFile, ", 128")
				bytesWritten += 1
			end if
			puts(outFile, CRLF)
		end for
		puts(outFile, ".globl " & name & "_tbl" & CRLF)
		puts(outFile, name & "_tbl:" & CRLF)
		for i = 1 to length(tbl[1]) do
			printf(outFile, ptrDecl & name & "_%d" & CRLF, tbl[1][i])
			bytesWritten += ptrSize
		end for
		puts(outFile, ptrDecl & "0" & CRLF)
		bytesWritten += 2
		if canLoop then
			puts(outFile, ".globl " & name & "_loop_tbl" & CRLF)
			puts(outFile, name & "_loop_tbl:" & CRLF)
			for i = 1 to length(tbl[1]) do
				printf(outFile, ptrDecl & name & "_%d_loop" & CRLF, tbl[1][i])
				bytesWritten += ptrSize
			end for
			puts(outFile, ptrDecl & "0" & CRLF)
			bytesWritten += ptrSize
		end if
		puts(outFile, {13, 10})
	else
		puts(outFile, ".globl " & name & "_tbl" & {13, 10})
		puts(outFile, name & "_tbl:" & {13, 10})
		if canLoop then
			puts(outFile, ".globl " & name & "_loop_tbl" & CRLF)
			puts(outFile, name & "_loop_tbl:" & CRLF)
		end if
		puts(outFile, ptrDecl & "0" & CRLF)
		bytesWritten += ptrSize
		puts(outFile, CRLF)
	end if
	return bytesWritten
end function


global function output_c_table(sequence name, sequence tbl, integer canLoop)
	integer bytesWritten
	
	bytesWritten = 0
	
	--? assoc_get_references(tbl)
	
	if length(tbl[1]) then
		for i = 1 to length(tbl[1]) do
			printf(outFile, "const unsigned char " & name & "_%d[] =" & {13, 10} & "{" & {13, 10}, tbl[1][i])
			for j = 1 to length(tbl[2][i][2]) do
				if canLoop and j = length(tbl[2][i][2]) and length(tbl[2][i][3]) = 0 then
					if j > 1 then
						puts(outFile, ",128")
						bytesWritten += 1
					end if
					printf(outFile, "};" & {13, 10} & "const unsigned char " & name & "_%d_loop[] =" & {13, 10} & "{" & {13, 10}, tbl[1][i])
					printf(outFile, "%d,128" & CRLF & "};", tbl[2][i][2][j])
					bytesWritten += 1
				elsif j = 1 then
					printf(outFile, "%d", tbl[2][i][2][j])
					bytesWritten += 1
				else
					if remainder(j,16)=0 then
						puts(outFile, CRLF)
					end if
					printf(outFile, ",%d", tbl[2][i][2][j])
					bytesWritten += 1
				end if
			end for
			if canLoop and length(tbl[2][i][3]) then
				if length(tbl[2][i][2]) then
					puts(outFile, ",128" & CRLF & "};")
					bytesWritten += 1
				end if
				printf(outFile, {13, 10} & "const unsigned char " & name & "_%d_loop[] =" & {13, 10} & "{" & {13, 10}, tbl[1][i])
				for j = 1 to length(tbl[2][i][3]) do
					if j = 1 then
						printf(outFile, "%d", tbl[2][i][3][j])
						bytesWritten += 1
					else
						if remainder(j,16)=0 then
							puts(outFile, CRLF)
						end if
						printf(outFile, ",%d", tbl[2][i][3][j])
						bytesWritten += 1
					end if
				end for
				puts(outFile, ",128" & CRLF & "};")
				bytesWritten += 1
			end if
			puts(outFile, {13, 10})
		end for
		puts(outFile, "const unsigned char *" & name & "_tbl[] =" & {13, 10} & "{" & {13, 10})
		for i = 1 to length(tbl[1]) do
			printf(outFile, name & "_%d", tbl[1][i])
			bytesWritten += 4
			if i < length(tbl[1]) then
				puts(outFile, ",")
			end if
			puts(outFile, {13, 10})
		end for
		puts(outFile, "};" & {13, 10})
		if canLoop then
			puts(outFile, "const unsigned char *" & name & "_loop_tbl[] = " & {13, 10} & "{" & {13, 10})
			for i = 1 to length(tbl[1]) do
				printf(outFile, name & "_%d_loop", tbl[1][i])
				bytesWritten += 4
				if i < length(tbl[1]) then
					puts(outFile, ",")
				end if
				puts(outFile, {13, 10})
			end for
			puts(outFile, "};")
		end if
		puts(outFile, {13, 10})
	else
		puts(outFile, "const unsigned char **" & name & "_tbl;" & {13, 10})
		if canLoop then
			puts(outFile, "const unsigned char **" & name & "_loop_tbl;" & {13, 10})
		end if
		puts(outFile, {13, 10})
	end if
	return bytesWritten
end function



-- Output data suitable for the SEGA Master System / Game Gear playback library (WLA-DX)
global procedure output_sms(integer writeVGM)
	atom factor
	integer f, machineSpeed, tableSize, cbSize, patSize, songSize, numSongs
	sequence freqTbl, oct1, fileEnding, s
	
	if writeVGM then
		fileEnding = ".vgm"
	else
		fileEnding = ".asm"
	end if

	numSongs = 0
	for i = 1 to length(songs) do
		if sequence(songs[i]) then
			numSongs += 1
		end if
	end for
	
	if writeVGM then
		if numSongs = 1 then
			write_vgm(shortFilename & fileEnding, 1, PSG_ENABLED, YM2151_DISABLED, supportsPAL, YM2612_DISABLED)
		else
			for i = 1 to length(songs) do
				if sequence(songs[i]) then
					write_vgm(shortFilename & sprintf("_song%d", i) & fileEnding, i, PSG_ENABLED, YM2151_DISABLED, supportsPAL, YM2612_DISABLED)
				end if
			end for
		end if
	else
		outFile = open(shortFilename & fileEnding, "wb")
		if outFile = -1 then
			ERROR("Unable to open file: " & shortFilename & fileEnding, -1)
		end if

		s = date()
		printf(outFile, "; Written by XPMC at %02d:%02d:%02d on " & WEEKDAYS[s[7]] & " " & MONTHS[s[2]] & " %d, %d." & {13, 10, 13, 10},
		       s[4..6] & {s[3], s[1] + 1900})
		       
		if not supportsPAL then
			puts(outFile, ".DEFINE XPMP_GAME_GEAR" & {13, 10})
		end if

		if updateFreq = 50 then
			puts(outFile, ".DEFINE XPMP_50_HZ" & {13, 10})
			machineSpeed = 3546893
		else
			machineSpeed = 3579545
		end if

		if target = TARGET_SMS and tune then
			puts(outFile, ".DEFINE XPMP_TUNE_SMS" & CRLF)
		end if

		if sum(assoc_get_references(volumeMacros)) = 0 then --not length(volumeMacros[1]) then
			puts(outFile, ".DEFINE XPMP_VMAC_NOT_USED" & CRLF)
		end if
		if sum(assoc_get_references(pitchMacros)) = 0 then
			puts(outFile, ".DEFINE XPMP_EPMAC_NOT_USED" & CRLF)
		end if
		if sum(assoc_get_references(vibratos)) = 0 then
			puts(outFile, ".DEFINE XPMP_MPMAC_NOT_USED" & CRLF)
		end if
		if not usesEN[1] then
			puts(outFile, ".DEFINE XPMP_ENMAC_NOT_USED" & CRLF)
		end if
		if not usesEN[2] then
			puts(outFile, ".DEFINE XPMP_EN2MAC_NOT_USED" & CRLF)
		end if

		for i = 1 to length(supportedChannels)-1 do
			for j = 1 to length(usesEffect[i]) do
				if usesEffect[i][j] then
					printf(outFile, ".DEFINE XPMP_CHN%d_USES_%s" & CRLF, {i - 1, EFFECT_STRINGS[j]})
				end if
			end for
		end for
				
		-- Write frequency table
		--factor = 2.0
		--for i = 2 to 7 do
		--	puts(outFile, ".dw ")
		--	for n = 1 to 12 do
		--		f = floor(machineSpeed / (OCTAVE1_ALT_SMS[n] * factor * 32))
		--		printf(outFile, "$%04x", and_bits(f, #3F0) * 16 + and_bits(f, #0F) + #80)
		--		if n < 12 then
		--			puts(outFile, ",")
		--		end if
		--	end for
		--	puts(outFile, {13,10})
		--	factor *= 2.0
		--end for
		--puts(outFile, {13, 10})

		tableSize  = output_wla_table("xpmp_dt_mac", dutyMacros,   1, 1, #80)
		tableSize += output_wla_table("xpmp_v_mac",  volumeMacros, 1, 1, #80)
		tableSize += output_wla_table("xpmp_VS_mac", volumeSlides, 1, 1, #80)
		tableSize += output_wla_table("xpmp_EP_mac", pitchMacros,  1, 1, #80)
		tableSize += output_wla_table("xpmp_EN_mac", arpeggios,    1, 1, #80)
		tableSize += output_wla_table("xpmp_MP_mac", vibratos,     0, 1, #80)
		tableSize += output_wla_table("xpmp_CS_mac", panMacros,    1, 1, #80)

		if verbose then
			printf(1, "Size of effect tables: %d bytes\n", tableSize)
		end if

		cbSize = 0
		puts(outFile, "xpmp_callback_tbl:" & {13, 10})
		for i = 1 to length(callbacks) do
			puts(outFile, ".dw " & callbacks[i] & {13, 10})
			cbSize += 2
		end for
		puts(outFile, {13, 10})

		if verbose then
			printf(1, "Size of callback table: %d bytes\n", cbSize)
		end if

		patSize = 0
		for n = 1 to length(patterns[2]) do
			printf(outFile, "xpmp_pattern%d:", n)
			for j = 1 to length(patterns[2][n]) do
				if remainder(j, 16) = 1 then
					puts(outFile, CRLF & ".db ")
				end if				
				printf(outFile, "$%02x", and_bits(patterns[2][n][j], #FF))
				if j < length(patterns[2][n]) and remainder(j, 16) != 0 then
					puts(outFile, ",")
				end if

			end for
			puts(outFile, CRLF)
			patSize += length(patterns[2][n])
		end for

		puts(outFile, {13, 10} & "xpmp_pattern_tbl:" & CRLF)
		for n = 1 to length(patterns[2]) do
			printf(outFile, ".dw xpmp_pattern%d" & {13, 10}, n)
			patSize += 2
		end for
		puts(outFile, {13, 10})
		
		if verbose then
			printf(1, "Size of patterns table: %d bytes\n", patSize)
		end if

		songSize = 0
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels)-1 do
					printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
					for j = 1 to length(songs[n][i]) do
						if remainder(j, 16) = 1 then
							puts(outFile, CRLF & ".db ")
						end if				
						printf(outFile, "$%02x", and_bits(songs[n][i][j], #FF))
						if j < length(songs[n][i]) and remainder(j, 16) != 0 then
							puts(outFile, ",")
						end if

					end for
					puts(outFile, CRLF)
					printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), -floor(-songLen[n][i]), -floor(-songLoopLen[n][i])})
					songSize += length(songs[n][i])
				end for
			end if
		end for

		puts(outFile, {13, 10} & "xpmp_song_tbl:" & CRLF)
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels)-1 do
					printf(outFile, ".dw xpmp_s%d_channel_" & supportedChannels[i] & {13, 10}, n)
					songSize += 2
				end for
			end if
		end for

		if verbose then
			printf(1, "Total size of song(s): %d bytes\n", songSize + patSize + tableSize + cbSize)
		end if

		close(outFile)
	end if
end procedure



-- Output data suitable for the KSS playback library (WLA-DX)
global procedure output_kss()
	atom factor
	integer f, machineSpeed, tableSize, cbSize, patSize, wavSize, songSize, numSongs, extraChips, usesPSG, usesSCC
	sequence freqTbl, oct1, fileEnding, s
	
	fileEnding = ".asm"

	extraChips = 0
	usesSCC = 0
	usesPSG = 0
	
	numSongs = 0
	for i = 1 to length(songs) do
		if sequence(songs[i]) then
			numSongs += 1
			if (length(songs[i][1]) + length(songs[i][2]) +
			    length(songs[i][3]) + length(songs[i][4])) != 4 then
				extraChips = or_bits(extraChips, 6)
			end if
			if (length(songs[i][5]) + length(songs[i][6]) +
			    length(songs[i][7])) != 3 then
				usesPSG = or_bits(usesPSG, 1)
			end if
			if (length(songs[i][8]) + length(songs[i][9]) +
			    length(songs[i][10]) + length(songs[i][11])) != 4 then
				usesSCC = or_bits(usesSCC, 1)
			end if
		end if
	end for
	
	outFile = open(shortFilename & fileEnding, "wb")
	if outFile = -1 then
		ERROR("Unable to open file: " & shortFilename & fileEnding, -1)
	end if
	s = date()
	printf(outFile, "; Written by XPMC at %02d:%02d:%02d on " & WEEKDAYS[s[7]] & " " & MONTHS[s[2]] & " %d, %d." & {13, 10, 13, 10},
	       s[4..6] & {s[3], s[1] + 1900})
	       
	--if updateFreq = 50 then
	--	puts(outFile, ".DEFINE XPMP_50_HZ" & {13, 10})
	--	machineSpeed = 3546893
	--else
		machineSpeed = 3579545
	--end if


	puts(outFile, 
		".IFDEF XPMP_MAKE_KSS" & CRLF &
		".memorymap" & CRLF &
   		"defaultslot 0" & CRLF &
   		"slotsize $8010" & CRLF &
   		"slot 0 0" & CRLF &
		".endme" & CRLF & CRLF &
		".rombanksize $8010" & CRLF &
		".rombanks 1" & CRLF & CRLF &
   		".orga   $0000" & CRLF &   
   		".db      \"KSCC\"   ; Magic string" & CRLF &
   		".dw   $0000   ; Load address" & CRLF &
   		".dw   $8000   ; Data length" & CRLF &
   		".dw   $7FF0   ; Driver initialize function" & CRLF &
   		".dw   $0000   ; Play address" & CRLF &
   		".db   $00   ; No. of banks" & CRLF &
   		".db   $00   ; extra" & CRLF &
   		".db   $00   ; reserved" & CRLF)
   	printf(outFile,
   		".db   $%02x   ; Extra chips" & CRLF & CRLF &
		".incbin \"kss.bin\"" & CRLF & CRLF &
		".ELSE" & CRLF & CRLF, extraChips)

	if sum(assoc_get_references(volumeMacros)) = 0 then --not length(volumeMacros[1]) then
		puts(outFile, ".DEFINE XPMP_VMAC_NOT_USED" & CRLF)
	end if
	if sum(assoc_get_references(pitchMacros)) = 0 then
		puts(outFile, ".DEFINE XPMP_EPMAC_NOT_USED" & CRLF)
	end if
	if sum(assoc_get_references(vibratos)) = 0 then
		puts(outFile, ".DEFINE XPMP_MPMAC_NOT_USED" & CRLF)
	end if
	if not usesEN[1] then
		puts(outFile, ".DEFINE XPMP_ENMAC_NOT_USED" & CRLF)
	end if
	if not usesEN[2] then
		puts(outFile, ".DEFINE XPMP_EN2MAC_NOT_USED" & CRLF)
	end if
	if usesPSG then
		puts(outFile, ".DEFINE XPMP_USES_AY" & CRLF)
	end if
	if usesSCC then
		puts(outFile, ".DEFINE XPMP_USES_SCC" & CRLF)
	end if
	if and_bits(extraChips, 2) then
		puts(outFile, ".DEFINE XPMP_USES_SN76489" & CRLF)
	end if
	
	for i = 1 to length(supportedChannels)-1 do
		for j = 1 to length(usesEffect[i]) do
			if usesEffect[i][j] then
				printf(outFile, ".DEFINE XPMP_CHN%d_USES_%s" & CRLF, {i - 1, EFFECT_STRINGS[j]})
			end if
		end for
	end for

	tableSize  = output_wla_table("xpmp_dt_mac", dutyMacros,   1, 1, #80)
	tableSize += output_wla_table("xpmp_v_mac",  volumeMacros, 1, 1, #80)
	tableSize += output_wla_table("xpmp_VS_mac", volumeSlides, 1, 1, #80)
	tableSize += output_wla_table("xpmp_EP_mac", pitchMacros,  1, 1, #80)
	tableSize += output_wla_table("xpmp_EN_mac", arpeggios,    1, 1, #80)
	tableSize += output_wla_table("xpmp_MP_mac", vibratos,     0, 1, #80)
	tableSize += output_wla_table("xpmp_CS_mac", panMacros,    1, 1, #80)

	if verbose then
		printf(1, "Size of effect tables: %d bytes\n", tableSize)
	end if

	wavSize = 0
	puts(outFile, "xpmp_waveform_data:")
	for i = 1 to length(waveforms[1]) do
		for j = 1 to length(waveforms[2][i][2]) do
			if j = 1 then
				puts(outFile, CRLF & ".db ")
			end if				
			printf(outFile, "$%02x", and_bits(waveforms[2][i][2][j],255))
			wavSize += 1
			if j < length(waveforms[2][i][2]) then
				puts(outFile, ",")
			end if

		end for
	end for
	puts(outFile, {13, 10, 13, 10})
	
	cbSize = 0
	puts(outFile, "xpmp_callback_tbl:" & {13, 10})
	for i = 1 to length(callbacks) do
		puts(outFile, ".dw " & callbacks[i] & {13, 10})
		cbSize += 2
	end for
	puts(outFile, {13, 10})

	if verbose then
		printf(1, "Size of callback table: %d bytes\n", cbSize)
	end if

	patSize = 0
	for n = 1 to length(patterns[2]) do
		printf(outFile, "xpmp_pattern%d:", n)
		for j = 1 to length(patterns[2][n]) do
			if remainder(j, 16) = 1 then
				puts(outFile, CRLF & ".db ")
			end if				
			printf(outFile, "$%02x", and_bits(patterns[2][n][j], #FF))
			if j < length(patterns[2][n]) and remainder(j, 16) != 0 then
				puts(outFile, ",")
			end if
			end for
		puts(outFile, CRLF)
		patSize += length(patterns[2][n])
	end for

	puts(outFile, {13, 10} & "xpmp_pattern_tbl:" & CRLF)
	for n = 1 to length(patterns[2]) do
		printf(outFile, ".dw xpmp_pattern%d" & {13, 10}, n)
		patSize += 2
	end for
	puts(outFile, {13, 10})
		
	if verbose then
		printf(1, "Size of patterns table: %d bytes\n", patSize)
	end if

	songSize = 0
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels)-1 do
				printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
				for j = 1 to length(songs[n][i]) do
					if remainder(j, 16) = 1 then
						puts(outFile, CRLF & ".db ")
					end if				
					printf(outFile, "$%02x", and_bits(songs[n][i][j], #FF))
					if j < length(songs[n][i]) and remainder(j, 16) != 0 then
						puts(outFile, ",")
					end if
					end for
				puts(outFile, CRLF)
				printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), -floor(-songLen[n][i]), -floor(-songLoopLen[n][i])})
				songSize += length(songs[n][i])
			end for
		end if
	end for

	puts(outFile, {13, 10} & "xpmp_song_tbl:" & CRLF)
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels)-1 do
				printf(outFile, ".dw xpmp_s%d_channel_" & supportedChannels[i] & {13, 10}, n)
				songSize += 2
			end for
		end if
	end for

	puts(outFile, ".ENDIF" & CRLF)
	
	if verbose then
		printf(1, "Total size of song(s): %d bytes\n", songSize + patSize + wavSize + tableSize + cbSize)
	end if

	close(outFile)
end procedure



-- Output data suitable for the ColecoVision playback library (WLA-DX)
global procedure output_clv(integer writeVGM)
	atom factor
	integer f, machineSpeed, tableSize, cbSize, patSize, songSize, numSongs
	sequence freqTbl, oct1, fileEnding, s
	
	if writeVGM then
		fileEnding = ".vgm"
	else
		fileEnding = ".asm"
	end if

	numSongs = 0
	for i = 1 to length(songs) do
		if sequence(songs[i]) then
			numSongs += 1
		end if
	end for
	
	if writeVGM then
		--if numSongs = 1 then
		--	write_vgm(shortFilename & fileEnding, 1, PSG_ENABLED, YM2151_DISABLED, supportsPAL, YM2612_DISABLED)
		--else
		--	for i = 1 to length(songs) do
		--		if sequence(songs[i]) then
		--			write_vgm(shortFilename & sprintf("_song%d", i) & fileEnding, i, PSG_ENABLED, YM2151_DISABLED, supportsPAL, YM2612_DISABLED)
		--		end if
		--	end for
		--end if
	else
		outFile = open(shortFilename & fileEnding, "wb")
		if outFile = -1 then
			ERROR("Unable to open file: " & shortFilename & fileEnding, -1)
		end if

		s = date()
		printf(outFile, "; Written by XPMC at %02d:%02d:%02d on " & WEEKDAYS[s[7]] & " " & MONTHS[s[2]] & " %d, %d." & {13, 10, 13, 10},
		       s[4..6] & {s[3], s[1] + 1900})
		       
		puts(outFile, ".DEFINE XPMP_COLECOVISION" & CRLF)

		if updateFreq = 50 then
			puts(outFile, ".DEFINE XPMP_50_HZ" & CRLF)
			machineSpeed = 3546893
		else
			machineSpeed = 3579545
		end if

		if target = TARGET_SMS and tune then
			puts(outFile, ".DEFINE XPMP_TUNE_SMS" & CRLF)
		end if

		if sum(assoc_get_references(volumeMacros)) = 0 then --not length(volumeMacros[1]) then
			puts(outFile, ".DEFINE XPMP_VMAC_NOT_USED" & CRLF)
		end if
		if sum(assoc_get_references(pitchMacros)) = 0 then
			puts(outFile, ".DEFINE XPMP_EPMAC_NOT_USED" & CRLF)
		end if
		if sum(assoc_get_references(vibratos)) = 0 then
			puts(outFile, ".DEFINE XPMP_MPMAC_NOT_USED" & CRLF)
		end if
		if not usesEN[1] then
			puts(outFile, ".DEFINE XPMP_ENMAC_NOT_USED" & CRLF)
		end if
		if not usesEN[2] then
			puts(outFile, ".DEFINE XPMP_EN2MAC_NOT_USED" & CRLF)
		end if

		for i = 1 to length(supportedChannels)-1 do
			for j = 1 to length(usesEffect[i]) do
				if usesEffect[i][j] then
					printf(outFile, ".DEFINE XPMP_CHN%d_USES_%s" & CRLF, {i - 1, EFFECT_STRINGS[j]})
				end if
			end for
		end for
		
		tableSize  = output_wla_table("xpmp_dt_mac", dutyMacros,   1, 1, #80)
		tableSize += output_wla_table("xpmp_v_mac",  volumeMacros, 1, 1, #80)
		tableSize += output_wla_table("xpmp_VS_mac", volumeSlides, 1, 1, #80)
		tableSize += output_wla_table("xpmp_EP_mac", pitchMacros,  1, 1, #80)
		tableSize += output_wla_table("xpmp_EN_mac", arpeggios,    1, 1, #80)
		tableSize += output_wla_table("xpmp_MP_mac", vibratos,     0, 1, #80)
		tableSize += output_wla_table("xpmp_CS_mac", panMacros,    1, 1, #80)

		if verbose then
			printf(1, "Size of effect tables: %d bytes\n", tableSize)
		end if

		cbSize = 0
		puts(outFile, "xpmp_callback_tbl:" & {13, 10})
		for i = 1 to length(callbacks) do
			puts(outFile, ".dw " & callbacks[i] & {13, 10})
			cbSize += 2
		end for
		puts(outFile, {13, 10})

		if verbose then
			printf(1, "Size of callback table: %d bytes\n", cbSize)
		end if

		patSize = 0
		for n = 1 to length(patterns[2]) do
			printf(outFile, "xpmp_pattern%d:", n)
			for j = 1 to length(patterns[2][n]) do
				if remainder(j, 16) = 1 then
					puts(outFile, CRLF & ".db ")
				end if				
				printf(outFile, "$%02x", and_bits(patterns[2][n][j], #FF))
				if j < length(patterns[2][n]) and remainder(j, 16) != 0 then
					puts(outFile, ",")
				end if

			end for
			puts(outFile, CRLF)
			patSize += length(patterns[2][n])
		end for

		puts(outFile, {13, 10} & "xpmp_pattern_tbl:" & CRLF)
		for n = 1 to length(patterns[2]) do
			printf(outFile, ".dw xpmp_pattern%d" & {13, 10}, n)
			patSize += 2
		end for
		puts(outFile, {13, 10})
		
		if verbose then
			printf(1, "Size of patterns table: %d bytes\n", patSize)
		end if

		songSize = 0
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels)-1 do
					printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
					for j = 1 to length(songs[n][i]) do
						if remainder(j, 16) = 1 then
							puts(outFile, CRLF & ".db ")
						end if				
						printf(outFile, "$%02x", and_bits(songs[n][i][j], #FF))
						if j < length(songs[n][i]) and remainder(j, 16) != 0 then
							puts(outFile, ",")
						end if

					end for
					puts(outFile, CRLF)
					printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), -floor(-songLen[n][i]), -floor(-songLoopLen[n][i])})
					songSize += length(songs[n][i])
				end for
			end if
		end for

		puts(outFile, {13, 10} & "xpmp_song_tbl:" & CRLF)
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels)-1 do
					printf(outFile, ".dw xpmp_s%d_channel_" & supportedChannels[i] & {13, 10}, n)
					songSize += 2
				end for
			end if
		end for

		if verbose then
			printf(1, "Total size of song(s): %d bytes\n", songSize + patSize + tableSize + cbSize)
		end if

		close(outFile)
	end if
end procedure


-- Output data suitable for the SEGA Genesis (Megadrive) playback library
global procedure output_gen(integer writeVGM)
	atom factor
	integer f, machineSpeed, tableSize, cbSize, songSize, patSize, numSongs
	sequence freqTbl, oct1, fileEnding, s, e
	
	if writeVGM then
		fileEnding = ".vgm"
	else
		fileEnding = ".s"
	end if

	-- Convert ADSR envelopes to the format used by the YM2612
	for i = 1 to length(adsrs[2]) do
		s = {0, 0, 0, 0}
		e = adsrs[2][i][2]
		s[1] = e[1]
		s[2] = e[2]
		s[3] = e[3]
		s[4] = floor(e[5] / 2) + floor(xor_bits(e[4], 31) / 2) * #10
		adsrs[2][i][2] = s
	end for
	
	for i = 1 to length(mods[2]) do
		s = mods[2][i][2]
		s[2] = s[2] * 8 + s[3]
		mods[2][i][2] = s[1..2]
	end for
	
	for i = 1 to length(feedbackMacros[1]) do
		feedbackMacros[2][i][2] = (feedbackMacros[2][i][2])*8
		feedbackMacros[2][i][3] = (feedbackMacros[2][i][3])*8
	end for
	
	numSongs = 0
	for i = 1 to length(songs) do
		if sequence(songs[i]) then
			numSongs += 1
		end if
	end for
	
	if writeVGM then
		if numSongs = 1 then
			write_vgm(shortFilename & fileEnding, 1, PSG_ENABLED, YM2151_DISABLED, YM2413_DISABLED, YM2612_ENABLED)
		else
			for i = 1 to length(songs) do
				if sequence(songs[i]) then
					write_vgm(shortFilename & sprintf("_song%d", i) & fileEnding, i, PSG_ENABLED, YM2151_DISABLED, YM2413_DISABLED, YM2612_ENABLED)
				end if
			end for
		end if
	else
		outFile = open(shortFilename & fileEnding, "wb")
		if outFile = -1 then
			ERROR("Unable to open file: " & shortFilename & fileEnding, -1)
		end if

		s = date()
		printf(outFile, "# Written by XPMC at %02d:%02d:%02d on " & WEEKDAYS[s[7]] & " " & MONTHS[s[2]] & " %d, %d." & CRLF & CRLF,
		       s[4..6] & {s[3], s[1] + 1900})
		       
		if updateFreq = 50 then
			puts(outFile, ".equ XPMP_50_HZ, 1" & CRLF)
			machineSpeed = 3546893
		else
			machineSpeed = 3579545
		end if
		
		tableSize  = output_m68kas_table("xpmp_dt_mac", dutyMacros,   1, 1, 0)
		tableSize += output_m68kas_table("xpmp_v_mac",  volumeMacros, 1, 1, 0)
		tableSize += output_m68kas_table("xpmp_VS_mac", volumeSlides, 1, 1, 0)
		tableSize += output_m68kas_table("xpmp_EP_mac", pitchMacros,  1, 1, 0)
		tableSize += output_m68kas_table("xpmp_EN_mac", arpeggios,    1, 1, 0)
		tableSize += output_m68kas_table("xpmp_FB_mac", feedbackMacros,1, 1, 0)
		tableSize += output_m68kas_table("xpmp_MP_mac", vibratos,     0, 1, 0)
		tableSize += output_m68kas_table("xpmp_CS_mac", panMacros,    1, 1, 0)
		tableSize += output_m68kas_table("xpmp_ADSR",   adsrs,        0, 1, 0)
		tableSize += output_m68kas_table("xpmp_MOD",    mods,         0, 1, 0)

		cbSize = 0
		
		if verbose then
			printf(1, "Size of effect tables: %d bytes\n", tableSize)
		end if

		patSize = 0
		for n = 1 to length(patterns[2]) do
			printf(outFile, "xpmp_pattern%d:", n)
			for j = 1 to length(patterns[2][n]) do
				if remainder(j, 16) = 1 then
					puts(outFile, CRLF & "dc.b ")
				end if				
				printf(outFile, "0x%02x", and_bits(patterns[2][n][j], #FF))
				if j < length(patterns[2][n]) and remainder(j, 16) != 0 then
					puts(outFile, ",")
				end if

			end for
			puts(outFile, CRLF)
			patSize += length(patterns[2][n])
		end for

		puts(outFile, CRLF & ".globl xpmp_pattern_tbl" & CRLF)
		puts(outFile, "xpmp_pattern_tbl:" & CRLF)
		for n = 1 to length(patterns[2]) do
			printf(outFile, "dc.w xpmp_pattern%d" & {13, 10}, n)
			patSize += 4
		end for
		puts(outFile, {13, 10})
		
		if verbose then
			printf(1, "Size of patterns table: %d bytes\n", patSize)
		end if
		
		songSize = 0
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels)-1 do
					printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
					for j = 1 to length(songs[n][i]) do
						if remainder(j, 16) = 1 then
							puts(outFile, CRLF & "dc.b ")
						end if				
						printf(outFile, "0x%02x", and_bits(songs[n][i][j], #FF))
						if j < length(songs[n][i]) and remainder(j, 16) != 0 then
							puts(outFile, ",")
						end if

					end for
					puts(outFile, CRLF)
					printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), -floor(-songLen[n][i]), -floor(-songLoopLen[n][i])})
					songSize += length(songs[n][i])
				end for
			end if
		end for

		puts(outFile, CRLF & ".globl xpmp_song_tbl")
		puts(outFile, CRLF & "xpmp_song_tbl:" & CRLF)
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels)-1 do
					printf(outFile, "dc.w xpmp_s%d_channel_" & supportedChannels[i] & CRLF, n)
					songSize += 4
				end for
			end if
		end for
		puts(outFile, "dc.w 0" & CRLF)
		songSize += 4
		
		if verbose then
			printf(1, "Total size of song(s): %d bytes\n", songSize + patSize + tableSize + cbSize)
		end if
		
		close(outFile)
	end if
end procedure


-- Output data suitable for the Commodore 64 playback library (WLA-DX)
global procedure output_c64()
	atom factor
	integer f, machineSpeed, tableSize, cbSize, songSize, patSize, numSongs
	sequence freqTbl, oct1, fileEnding, e, s
	
	fileEnding = ".asm"

	-- Convert ADSR envelopes to the format used by the SID
	for i = 1 to length(adsrs[2]) do
		s = {0, 0}
		e = adsrs[2][i][2]
		s[1] = xor_bits(e[1], 15) * #10 + xor_bits(e[2], 15)
		s[2] = e[3] * #10 + xor_bits(e[4], 15)
		adsrs[2][i][2] = s
	end for
	
	for i = 1 to length(filters[2]) do
		s = filters[2][i][2]
		if s[1] = 0 then
			s[1] = #10
		elsif s[1] = 1 then
			s[1] = #40
		elsif s[1] = 2 then
			s[1] = #20
		else
			s[1] = #50
		end if
		s[3] *= #10
		s = {s[1], and_bits(s[2], 7), floor(s[2] / 8), s[3]}
		filters[2][i][2] = s
	end for
	
	for i = 1 to length(dutyMacros[2]) do
		for j = 1 to length(dutyMacros[2][i][2]) do
			dutyMacros[2][i][2][j] = power(2, dutyMacros[2][i][2][j]) * #10
		end for
		for j = 1 to length(dutyMacros[2][i][3]) do
			dutyMacros[2][i][3][j] = power(2, dutyMacros[2][i][3][j]) * #10
		end for
	end for
	
	numSongs = 0
	for i = 1 to length(songs) do
		if sequence(songs[i]) then
			numSongs += 1
		end if
	end for
	
	outFile = open(shortFilename & fileEnding, "wb")
	if outFile = -1 then
		ERROR("Unable to open file: " & shortFilename & fileEnding, -1)
	end if

	s = date()
	printf(outFile, "; Written by XPMC at %02d:%02d:%02d on " & WEEKDAYS[s[7]] & " " & MONTHS[s[2]] & " %d, %d." & {13, 10, 13, 10},
	       s[4..6] & {s[3], s[1] + 1900})

	puts(outFile,
	".IFDEF XPMP_MAKE_SID" & CRLF & CRLF &
	".MEMORYMAP" & CRLF &
	"\tDEFAULTSLOT 0"  & CRLF &
	"\tSLOTSIZE $8000"  & CRLF &
	"\tSLOT 0 $0000"  & CRLF &
	".ENDME"  & CRLF & CRLF)
	puts(outFile,
	".ROMBANKSIZE $8000"  & CRLF &
	".ROMBANKS 1"  & CRLF &
	".BANK 0 SLOT 0"  & CRLF &
	".ORGA $00"  & CRLF & CRLF)
	puts(outFile,
	".db \"PSID\""  & CRLF &
	".db $00,$01\t; Version"  & CRLF &
	".db $00,$76\t; Data address"  & CRLF &
	".dw 0\t\t; Load address (defined by first two bytes of psid.bin)"  & CRLF &
	".dw 0\t\t; Init address (same as load address)" & CRLF &
	".dw $1110\t; Play address" & CRLF &
	sprintf(".dw $%02x%02x\t; Songs",{and_bits(numSongs, #FF), floor(numSongs / #100)}) & CRLF &
	".dw $0100\t; Start song" & CRLF &
	".dw 0,0\t\t; Speed" & CRLF)
	if length(songTitle) >= 32 then
		puts(outFile, ".db \"" & songTitle[1..31] & "\", 0" & CRLF)
	else
		puts(outFile, ".db \"" & songTitle & "\"")
		for i = 1 to 32 - length(songTitle) do
			puts(outFile, ", 0")
		end for
		puts(outFile, CRLF)
	end if
	if length(songComposer) >= 32 then
		puts(outFile, ".db \"" & songComposer[1..31] & "\", 0" & CRLF)
	else
		puts(outFile, ".db \"" & songComposer & "\"")
		for i = 1 to 32 - length(songComposer) do
			puts(outFile, ", 0")
		end for
		puts(outFile, CRLF)
	end if
	s = songProgrammer & sprintf(", %d", s[1] + 1900)
	if length(s) >= 32 then
		puts(outFile, ".db \"" & s[1..31] & "\", 0" & CRLF)
	else
		puts(outFile, ".db \"" & s & "\"")
		for i = 1 to 32 - length(s) do
			puts(outFile, ", 0")
		end for
		puts(outFile, CRLF)
	end if
	puts(outFile,".INCBIN \"psid.bin\"" & CRLF)
	
	puts(outFile, CRLF & ".ELSE" & CRLF & CRLF)

	if updateFreq = 50 then
		puts(outFile, ".DEFINE XPMP_50_HZ" & {13, 10})
	else
		machineSpeed = 3579545
	end if

	if sum(assoc_get_references(volumeMacros)) = 0 then --not length(volumeMacros[1]) then
		puts(outFile, ".DEFINE XPMP_VMAC_NOT_USED" & CRLF)
	end if
	if sum(assoc_get_references(pitchMacros)) = 0 then
		puts(outFile, ".DEFINE XPMP_EPMAC_NOT_USED" & CRLF)
	end if
	if sum(assoc_get_references(dutyMacros)) = 0 then
		puts(outFile, ".DEFINE XPMP_DTMAC_NOT_USED" & CRLF)
	end if
	if sum(assoc_get_references(pulseMacros)) = 0 then
		puts(outFile, ".DEFINE XPMP_PWMAC_NOT_USED" & CRLF)
	end if
	if sum(assoc_get_references(vibratos)) = 0 then
		puts(outFile, ".DEFINE XPMP_MPMAC_NOT_USED" & CRLF)
	end if
	if not usesEN[1] then
		puts(outFile, ".DEFINE XPMP_ENMAC_NOT_USED" & CRLF)
	end if
	if not usesEN[2] then
		puts(outFile, ".DEFINE XPMP_EN2MAC_NOT_USED" & CRLF)
	end if

	for i = 1 to length(supportedChannels)-1 do
		for j = 1 to length(usesEffect[i]) do
			if usesEffect[i][j] then
				printf(outFile, ".DEFINE XPMP_CHN%d_USES_%s" & CRLF, {i - 1, EFFECT_STRINGS[j]})
			end if
		end for
	end for
		
	-- Write frequency table
	--factor = 0.5
	--for i = 0 to 7 do
	--	puts(outFile, ".dw ")
	--	for n = 1 to 12 do
	--		f = floor((OCTAVE1[n] * factor) / 0.06097)
	--		printf(outFile, "$%04x", f)
	--		if n < 12 then
	--			puts(outFile, ",")
	--		end if
	--	end for
	--	puts(outFile, {13,10})
	--	factor *= 2.0
	--end for
	--puts(outFile, {13, 10})

	tableSize  = output_wla_table("xpmp_dt_mac", dutyMacros,   1, 1, #00)
	tableSize += output_wla_table("xpmp_v_mac",  volumeMacros, 1, 1, #80)
	tableSize += output_wla_table("xpmp_VS_mac", volumeSlides, 1, 1, #80)
	tableSize += output_wla_table("xpmp_EP_mac", pitchMacros,  1, 1, #80)
	tableSize += output_wla_table("xpmp_PW_mac", pulseMacros,  1, 1, #80)
	tableSize += output_wla_table("xpmp_EN_mac", arpeggios,    1, 1, #80)
	tableSize += output_wla_table("xpmp_MP_mac", vibratos,     0, 1, #80)
	tableSize += output_wla_table("xpmp_CS_mac", panMacros,    1, 1, #80)
	--tableSize += output_wla_table("xpmp_ADSR",   adsrs,        0, 1, #80)
	tableSize += output_wla_table("xpmp_FT",     filters,      0, 1, #80)
	

	puts(outFile,"xpmp_ADSR_tbl:" & CRLF)

	for i = 1 to length(adsrs[2]) do
		printf(outFile, ".db $%02x, $%02x" & CRLF, adsrs[2][i][2])
	end for
	puts(outFile, CRLF)
	
	if verbose then
		printf(1, "Size of effect tables: %d bytes\n", tableSize)
	end if

	cbSize = 0
	puts(outFile, "xpmp_callback_tbl:" & {13, 10})
	for i = 1 to length(callbacks) do
		puts(outFile, ".dw " & callbacks[i] & {13, 10})
		cbSize += 2
	end for
	puts(outFile, {13, 10})

	if verbose then
		printf(1, "Size of callback table: %d bytes\n", cbSize)
	end if

	patSize = 0
	for n = 1 to length(patterns[2]) do
		printf(outFile, "xpmp_pattern%d:", n)
		for j = 1 to length(patterns[2][n]) do
			if remainder(j, 16) = 1 then
				puts(outFile, CRLF & ".db ")
			end if				
			printf(outFile, "$%02x", and_bits(patterns[2][n][j], #FF))
			if j < length(patterns[2][n]) and remainder(j, 16) != 0 then
				puts(outFile, ",")
			end if

		end for
		puts(outFile, CRLF)
		patSize += length(patterns[2][n])
	end for

	puts(outFile, {13, 10} & "xpmp_pattern_tbl:" & CRLF)
	for n = 1 to length(patterns[2]) do
		printf(outFile, ".dw xpmp_pattern%d" & {13, 10}, n)
		patSize += 2
	end for
	puts(outFile, {13, 10})

	if verbose then
		printf(1, "Size of patterns table: %d bytes\n", patSize)
	end if
		
	songSize = 0
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels)-1 do
				printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
				for j = 1 to length(songs[n][i]) do
					if remainder(j, 16) = 1 then
						puts(outFile, CRLF & ".db ")
					end if				
					printf(outFile, "$%02x", and_bits(songs[n][i][j], #FF))
					if j < length(songs[n][i]) and remainder(j, 16) != 0 then
						puts(outFile, ",")
					end if

				end for
				puts(outFile, CRLF)
				printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), -floor(-songLen[n][i]), -floor(-songLoopLen[n][i])})
				songSize += length(songs[n][i])
			end for
		end if
	end for

	puts(outFile, {13, 10} & "xpmp_song_tbl:" & CRLF)
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels)-1 do
				printf(outFile, ".dw xpmp_s%d_channel_" & supportedChannels[i] & CRLF, n)
				songSize += 2
			end for
		end if
	end for

	if verbose then
		printf(1, "Total size of song(s): %d bytes\n", songSize + patSize + tableSize + cbSize)
	end if

	puts(outFile, ".ENDIF" & CRLF)
	
	close(outFile)
end procedure


-- Output VGM data for the CPS-1 (YM2151)
global procedure output_cps(integer writeVGM)
	atom factor
	integer f, machineSpeed, tableSize, cbSize, songSize, numSongs
	sequence freqTbl, oct1, fileEnding, s, e

	-- Force a .vgm file ending since it's the only format supported	
	fileEnding = ".vgm"

	-- Convert ADSR envelopes to the format used by the YM2151
	for i = 1 to length(adsrs[2]) do
		s = {0, 0, 0, 0}
		e = adsrs[2][i][2]
		s[1] = e[1]
		s[2] = e[2]
		s[3] = e[3]
		s[4] = floor(e[5] / 2) + floor(xor_bits(e[4], 31) / 2) * #10
		adsrs[2][i][2] = s
	end for
	
	for i = 1 to length(mods[2]) do
		s = mods[2][i][2]
		--s[2] = s[2] * 8 + s[3]
		--mods[2][i][2] = s[1..2]
		s[3] = or_bits(s[3], #80)
		s[4] += s[5] * #10
		s[5] = s[6] + #C0
		mods[2][i][2] = s[1..5]
	end for
	
	for i = 1 to length(feedbackMacros[1]) do
		feedbackMacros[2][i][2] = (feedbackMacros[2][i][2])*8
		feedbackMacros[2][i][3] = (feedbackMacros[2][i][3])*8
	end for
	
	numSongs = 0
	for i = 1 to length(songs) do
		if sequence(songs[i]) then
			numSongs += 1
		end if
	end for
	
	if numSongs = 1 then
		write_vgm(shortFilename & fileEnding, 1, PSG_DISABLED, YM2151_ENABLED, YM2413_DISABLED, YM2612_DISABLED)
	else
		for i = 1 to length(songs) do
			if sequence(songs[i]) then
				write_vgm(shortFilename & sprintf("_song%d", i) & fileEnding, i, PSG_DISABLED, YM2151_ENABLED, YM2413_DISABLED, YM2612_DISABLED)
			end if
		end for
	end if
end procedure


-- Output MDX data for the X68000 (YM2151)
global procedure output_x68()
	atom factor
	integer f, machineSpeed, tableSize, cbSize, songSize, numSongs
	sequence freqTbl, oct1, fileEnding, s, e

	-- Force an .mdx file ending since it's the only format supported	
	fileEnding = ".mdx"

	-- Convert ADSR envelopes to the format used by the YM2151
	for i = 1 to length(adsrs[2]) do
		s = {0, 0, 0, 0}
		e = adsrs[2][i][2]
		s[1] = e[1]
		s[2] = e[2]
		s[3] = e[3]
		s[4] = floor(e[5] / 2) + floor(xor_bits(e[4], 31) / 2) * #10
		adsrs[2][i][2] = s
	end for
	
	for i = 1 to length(mods[2]) do
		s = mods[2][i][2]
		s[3] = or_bits(s[3], #80)
		s[4] += s[5] * #10
		s[5] = s[6] + #C0
		mods[2][i][2] = s[1..5]
	end for
	
	for i = 1 to length(feedbackMacros[1]) do
		feedbackMacros[2][i][2] = (feedbackMacros[2][i][2])*8
		feedbackMacros[2][i][3] = (feedbackMacros[2][i][3])*8
	end for
	
	numSongs = 0
	for i = 1 to length(songs) do
		if sequence(songs[i]) then
			numSongs += 1
		end if
	end for
	
	if numSongs = 1 then
	--	write_mdx(shortFilename & fileEnding, 1)
	else
		for i = 1 to length(songs) do
			if sequence(songs[i]) then
	--			write_mdx(shortFilename & sprintf("_song%d", i) & fileEnding, i)
			end if
		end for
	end if
end procedure


-- Output data suitable for the Gameboy / Gameboy Color
global procedure output_gbc()
	atom factor,f2,r2,smallestDiff
	integer f, tableSize, cbSize, songSize, wavSize, patSize, numSongs
	sequence closest,s

	outFile = open(shortFilename & ".asm", "wb")
	if outFile = -1 then
		ERROR("Unable to open file: " & shortFilename & ".asm", -1)
	end if

	s = date()
	printf(outFile, "; Written by XPMC at %02d:%02d:%02d on " & WEEKDAYS[s[7]] & " " & MONTHS[s[2]] & " %d, %d." & {13, 10, 13, 10},
	       s[4..6] & {s[3], s[1] + 1900})

	numSongs = 0
	for i = 1 to length(songs) do
		if sequence(songs[i]) then
			numSongs += 1
		end if
	end for
	
	puts(outFile,
	".IFDEF XPMP_MAKE_GBS" & CRLF & CRLF &
	".MEMORYMAP" & CRLF &
	"\tDEFAULTSLOT 1"  & CRLF &
	"\tSLOTSIZE $4000"  & CRLF &
	"\tSLOT 0 $0000"  & CRLF &
	"\tSLOT 1 $4000"  & CRLF &
	".ENDME"  & CRLF & CRLF)
	puts(outFile,
	".ROMBANKSIZE $4000"  & CRLF &
	".ROMBANKS 2"  & CRLF &
	".BANK 0 SLOT 0"  & CRLF &
	".ORGA $00"  & CRLF & CRLF)
	
	puts(outFile,
	".db \"GBS\"" & CRLF &
	".db 1\t\t; Version" & CRLF &
	sprintf(".db %d\t\t; Number of songs", numSongs) & CRLF &
	".db 1\t\t; Start song" & CRLF &
	".dw $0400\t; Load address" & CRLF &
	".dw $0400\t; Init address" & CRLF &
	".dw $0408\t; Play address" & CRLF &
	".dw $fffe\t; Stack pointer" & CRLF &
	".db 0" & CRLF &
	".db 0" & CRLF)
	
	if length(songTitle) >= 32 then
		puts(outFile, ".db \"" & songTitle[1..31] & "\", 0" & CRLF)
	else
		puts(outFile, ".db \"" & songTitle & "\"")
		for i = 1 to 32 - length(songTitle) do
			puts(outFile, ", 0")
		end for
		puts(outFile, CRLF)
	end if
	if length(songComposer) >= 32 then
		puts(outFile, ".db \"" & songComposer[1..31] & "\", 0" & CRLF)
	else
		puts(outFile, ".db \"" & songComposer & "\"")
		for i = 1 to 32 - length(songComposer) do
			puts(outFile, ", 0")
		end for
		puts(outFile, CRLF)
	end if
	if length(songProgrammer) >= 32 then
		puts(outFile, ".db \"" & songProgrammer[1..31] & "\", 0" & CRLF)
	else
		puts(outFile, ".db \"" & songProgrammer & "\"")
		for i = 1 to 32 - length(songProgrammer) do
			puts(outFile, ", 0")
		end for
		puts(outFile, CRLF)
	end if	
	puts(outFile, ".INCBIN \"gbs.bin\"" & CRLF & CRLF)
	puts(outFile, ".ELSE" & CRLF & CRLF) 


	if sum(assoc_get_references(volumeMacros)) = 0 then 
		puts(outFile, ".DEFINE XPMP_VMAC_NOT_USED" & CRLF)
	end if
	if sum(assoc_get_references(pitchMacros)) = 0 then
		puts(outFile, ".DEFINE XPMP_EPMAC_NOT_USED" & CRLF)
	end if
	if sum(assoc_get_references(vibratos)) = 0 then
		puts(outFile, ".DEFINE XPMP_MPMAC_NOT_USED" & CRLF)
	end if
	if not usesEN[1] then
		puts(outFile, ".DEFINE XPMP_ENMAC_NOT_USED" & CRLF)
	end if
	if not usesEN[2] then
		puts(outFile, ".DEFINE XPMP_EN2MAC_NOT_USED" & CRLF)
	end if

	for i = 1 to length(supportedChannels)-1 do
		for j = 1 to length(usesEffect[i]) do
			if usesEffect[i][j] then
				printf(outFile, ".DEFINE XPMP_CHN%d_USES_%s" & CRLF, {i - 1, EFFECT_STRINGS[j]})
			end if
		end for
	end for
		
	tableSize  = output_wla_table("xpmp_dt_mac", dutyMacros,   1, 1, #80)
	tableSize += output_wla_table("xpmp_v_mac",  volumeMacros, 1, 1, #80)
	tableSize += output_wla_table("xpmp_VS_mac", volumeSlides, 1, 1, #80)
	tableSize += output_wla_table("xpmp_EP_mac", pitchMacros,  1, 1, #80)
	tableSize += output_wla_table("xpmp_EN_mac", arpeggios,    1, 1, #80)
	tableSize += output_wla_table("xpmp_MP_mac", vibratos,     0, 1, #80)
	tableSize += output_wla_table("xpmp_CS_mac", panMacros,    1, 1, #80)
	
	wavSize = 0
	puts(outFile, "xpmp_waveform_data:")
	for i = 1 to length(waveforms[1]) do
		for j = 1 to length(waveforms[2][i][2]) by 2 do
			if j = 1 then
				puts(outFile, CRLF & ".db ")
			end if				
			printf(outFile, "$%02x", waveforms[2][i][2][j]*#10 + waveforms[2][i][2][j+1])
			wavSize += 1
			if j < length(waveforms[2][i][2]) - 1 then
				puts(outFile, ",")
			end if

		end for
	end for
	puts(outFile, {13, 10, 13, 10})
	
	cbSize = 0
	puts(outFile, "xpmp_callback_tbl:" & CRLF)
	for i = 1 to length(callbacks) do
		puts(outFile, ".dw " & callbacks[i] & CRLF)
		cbSize += 2
	end for
	puts(outFile, CRLF)

	if verbose then
		printf(1, "Size of effect tables: %d bytes\n", tableSize)
	end if
	if verbose then
		printf(1, "Size of waveform table: %d bytes\n", wavSize)
	end if

	patSize = 0
	for n = 1 to length(patterns[2]) do
		printf(outFile, "xpmp_pattern%d:", n)
		for j = 1 to length(patterns[2][n]) do
			if remainder(j, 16) = 1 then
				puts(outFile, CRLF & ".db ")
			end if				
			printf(outFile, "$%02x", and_bits(patterns[2][n][j], #FF))
			if j < length(patterns[2][n]) and remainder(j, 16) != 0 then
				puts(outFile, ",")
			end if

		end for
		puts(outFile, CRLF)
		patSize += length(patterns[2][n])
	end for

	puts(outFile, {13, 10} & "xpmp_pattern_tbl:" & CRLF)
	for n = 1 to length(patterns[2]) do
		printf(outFile, ".dw xpmp_pattern%d" & {13, 10}, n)
		patSize += 2
	end for
	puts(outFile, {13, 10})

	if verbose then
		printf(1, "Size of patterns table: %d bytes\n", patSize)
	end if
	
	songSize = 0
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels)-1 do
				printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
				for j = 1 to length(songs[n][i]) do
					if remainder(j, 16) = 1 then
						puts(outFile, CRLF & ".db ")
					end if				
					printf(outFile, "$%02x", and_bits(songs[n][i][j], #FF))
					songSize += 1
					if j < length(songs[n][i]) and remainder(j, 16) != 0 then
						puts(outFile, ",")
					end if

				end for
				puts(outFile, CRLF)
				printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), -floor(-songLen[n][i]), -floor(-songLoopLen[n][i])})
			end for
		end if
	end for
	
	puts(outFile, {13, 10} & "xpmp_song_tbl:" & CRLF)
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels)-1 do
				printf(outFile, ".dw xpmp_s%d_channel_" & supportedChannels[i] & CRLF, n)
				songSize += 2
			end for
		end if
	end for

	if verbose then
		printf(1, "Total size of song(s): %d bytes\n", songSize + patSize + tableSize + cbSize + wavSize)
	end if
	
	puts(outFile, ".ENDIF")
	close(outFile)
end procedure


global procedure output_gba()
	integer first, songSize, tableSize
	sequence s
	
	outFile = open(shortFilename & ".c", "wb")
	if outFile = -1 then
		ERROR("Unable to open file: " & shortFilename & ".c", -1)
	end if

	s = date()
	printf(outFile, "// Written by XPMC at %02d:%02d:%02d on " & WEEKDAYS[s[7]] & " " & MONTHS[s[2]] & " %d, %d." & {13, 10, 13, 10},
	       s[4..6] & {s[3], s[1] + 1900})

	tableSize = 0
	tableSize =  output_c_table("xpmp_dt_mac", dutyMacros,   1)
	tableSize += output_c_table("xpmp_v_mac",  volumeMacros, 1)
	tableSize += output_c_table("xpmp_p_mac",  panMacros,    1)
	tableSize += output_c_table("xpmp_VS_mac", volumeSlides, 1)
	tableSize += output_c_table("xpmp_EP_mac", pitchMacros,  1)
	tableSize += output_c_table("xpmp_EN_mac", arpeggios,    1)
	tableSize += output_c_table("xpmp_MP_mac", vibratos,     0)
	
	--puts(outFile, "const xpmp_callback xpmp_callback_tbl[] =" & {13, 10} & "{" & {13, 10})
	--for i = 1 to length(callbacks) do
	--	puts(outFile, callbacks[i])
	--	puts(outFile, ",")
	--	puts(outFile, {13, 10})
	--end for
	--puts(outFile, "NULL" & {13, 10} & "};" & {13, 10})
	
	songSize = 0

	if verbose then
		printf(1, "Size of effect tables: %d bytes\n", tableSize)
	end if
	
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels)-1 do
				printf(outFile, "const unsigned char xpmp_s%d_channel_" & supportedChannels[i] & "[] = " & {13,10} & "{" & {13, 10}, n)
				for j = 1 to length(songs[n][i]) do
					printf(outFile, "0x%02x", songs[n][i][j])
					songSize += 1
					if j < length(songs[n][i]) then
						puts(outFile, ",")
					end if
					if remainder(j, 16) = 0 then
						puts(outFile, {13, 10})
					end if
				end for
				puts(outFile, "};" & {13, 10})
				printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), -floor(-songLen[n][i]), -floor(-songLoopLen[n][i])})
			end for
		end if
	end for
	
	puts(outFile, {13, 10} & "unsigned char const *xpmp_song_tbl[] = " & {13, 10} & "{" & {13, 10})
	first = 1
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels)-1 do
				if first then
					first = 0
				else
					puts(outFile, "," & {13,10})
				end if
				printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i], n)
				songSize += 2
			end for
		end if
	end for
	puts(outFile, {13, 10} & "};" & {13, 10})

	if verbose then
		printf(1, "Total size of song(s): %d bytes\n", songSize + tableSize) -- + cbSize + wavSize)
	end if
	
	close(outFile)
end procedure


global procedure output_nds()
	integer first, songSize, tableSize
	
	outFile = open(shortFilename & ".c", "wb")
	if outFile = -1 then
		ERROR("Unable to open file: " & shortFilename & ".c", -1)
	end if

	tableSize = 0
	tableSize =  output_c_table("xpmp_dt_mac", dutyMacros,   1)
	tableSize += output_c_table("xpmp_v_mac",  volumeMacros, 1)
	tableSize += output_c_table("xpmp_p_mac",  panMacros,    1)
	tableSize += output_c_table("xpmp_VS_mac", volumeSlides, 1)
	tableSize += output_c_table("xpmp_EP_mac", pitchMacros,  1)
	tableSize += output_c_table("xpmp_EN_mac", arpeggios,    1)
	tableSize += output_c_table("xpmp_MP_mac", vibratos,     0)
	
	--puts(outFile, "const xpmp_callback xpmp_callback_tbl[] =" & {13, 10} & "{" & {13, 10})
	--for i = 1 to length(callbacks) do
	--	puts(outFile, callbacks[i])
	--	puts(outFile, ",")
	--	puts(outFile, {13, 10})
	--end for
	--puts(outFile, "NULL" & {13, 10} & "};" & {13, 10})

	if verbose then
		printf(1, "Size of effect tables: %d bytes\n", tableSize)
	end if
	
	songSize = 0
	
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels)-1 do
				printf(outFile, "const unsigned char xpmp_s%d_channel_" & supportedChannels[i] & "[] = " & {13,10} & "{" & {13, 10}, n)
				for j = 1 to length(songs[n][i]) do
					printf(outFile, "0x%02x", songs[n][i][j])
					songSize += 1
					if j < length(songs[n][i]) then
						puts(outFile, ",")
					end if
					if remainder(j, 16) = 0 then
						puts(outFile, {13, 10})
					end if
				end for
				puts(outFile, "};" & {13, 10})
				printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), -floor(-songLen[n][i]), -floor(-songLoopLen[n][i])})
			end for
		end if
	end for
	
	puts(outFile, {13, 10} & "unsigned char const *xpmp_song_tbl[] = " & {13, 10} & "{" & {13, 10})
	first = 1
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels)-1 do
				if first then
					first = 0
				else
					puts(outFile, "," & {13,10})
				end if
				printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i], n)
				songSize += 2
			end for
		end if
	end for
	puts(outFile, {13, 10} & "};" & {13, 10})

	if verbose then
		printf(1, "Total size of song(s): %d bytes\n", songSize + tableSize) -- + cbSize + wavSize)
	end if
	
	close(outFile)
end procedure



-- Output YM data for the Atari ST (YM2149)
global procedure output_ast(integer writeYM)
	atom factor
	integer f, machineSpeed, tableSize, cbSize, songSize, numSongs
	sequence freqTbl, oct1, fileEnding, s, e

	-- Force a .ym file ending since it's the only format supported	
	fileEnding = ".ym"

	numSongs = 0
	for i = 1 to length(songs) do
		if sequence(songs[i]) then
			numSongs += 1
		end if
	end for
	
	if numSongs = 1 then
--		write_ym(shortFilename & fileEnding, 1, 1773400)
	else
		for i = 1 to length(songs) do
			if sequence(songs[i]) then
--				write_ym(shortFilename & sprintf("_song%d", i) & fileEnding, i, 1773400)
			end if
		end for
	end if
end procedure



global procedure output_pc4(integer writeWAV)
	atom factor
	integer f, machineSpeed, tableSize, cbSize, songSize, numSongs
	sequence freqTbl, oct1, fileEnding, s, e

	-- Force a .wav file ending since it's the only format supported	
	fileEnding = ".wav"

	numSongs = 0
	for i = 1 to length(songs) do
		if sequence(songs[i]) then
			numSongs += 1
		end if
	end for
	
	if numSongs = 1 then
		--write_wav_4ksynth(shortFilename & fileEnding, 1)
	else
		for i = 1 to length(songs) do
			if sequence(songs[i]) then
				--write_wav_4ksynth(shortFilename & sprintf("_song%d", i) & fileEnding, i)
			end if
		end for
	end if
end procedure
